#include <math.h>
#include <LEDA/ugraph.h>
#include <LEDA/graph_alg.h>
#include <LEDA/array.h>
#include <LEDA/graphwin.h>
#include <iostream.h>
#include <stdlib.h>

#define _slot 0
#define _peg 1
#define _plus 1
#define _minus -1
#define _zero 0
#define min(a, b)	a>b?b:a


/* Data structure to keep track of the pieces */
class piece {
public:
	friend istream& operator>> (istream&, piece&);
   friend ostream& operator<< (ostream&, const piece&);

  	piece();

  	piece (const piece& P);
   piece& operator= (const piece& P);

   /* if more vars are added change constructors and overloaded ops!! */
	int type; //type of piece = _plus || _minus || _zero
   node start_node, end_node; //start and end node of piece
   double length; //length of piece
	int match_piece;  //the piece this piece is matched to
   node match_node;  //the node the match edge is connected to
   double match_length; //the length of the match edge
   int serviced; //has this piece been serviced?

};


/* Data structure to hold information for each edge of the bipartite graph*/
class B_edge {
public:
	friend istream& operator>> (istream&, B_edge&);
   friend ostream& operator<< (ostream&, const B_edge&);

  	B_edge();
  	B_edge (const B_edge& B);
   B_edge& operator= (const B_edge& B);

   /* if more vars are added change constructors and overloaded ops!! */
	node p_node, m_node; //adjacent nodes of edge in original graph
	double length;	//length of the edge
};

void calc_excess(UGRAPH<int, double>& T, edge_array<int>& excess,
	const node_map<int>& node_type);
void match (const UGRAPH<int, double>& T,
	const node_map<point>& node_coord,
	const list<int>& plusL, const list<int>& minusL,
   array<piece>& P);
double create_BG (const UGRAPH<int, double>& T, const node_map<point>& node_coord,
	const list<int>& plusL, const list<int>& minusL,
   const array<piece>& P, GRAPH<int, B_edge>& B);
void cut_edges (const int i, UGRAPH<int, double>& T, const edge_array<int> excess,
	list<edge>& L, const int k);
double ac_tour (const UGRAPH<int, double>& T,
 	const node_map<point>& node_coord,
 	array<piece> P,
   const node_array<int>& node2Pnum);
double c_tour (const UGRAPH<int, double>& T,
 	const node_map<point>& node_coord,
 	array<piece> P,
   const node_array<int>& node2Pnum);
void get_pieces (const UGRAPH<int, double>& T, const list<edge>& L,
	const edge_array<int> excess,
	array<piece>& P, const node_array<int>& node2Pnum, list<int>& plusL,
   list<int>& minusL);
double TSPtour_len (const UGRAPH<int, double>& T,
 	const node_map<point>& node_coord);
double dist (const node a, const node b, const node_map<point>& node_coord)
//calculates distance between 2 nodes
{
        return (node_coord[a].distance(node_coord[b]));
}


/*******************************************************************************
FUNCTION ckr

Input:
Output: length of tour. if length = tsp length then -1.

Graph T has to be a TSP solution with the nodes and edges created
in the order of the traversal.
*******************************************************************************/
double ckr(UGRAPH<int, double>& T,
	node_map<point>& node_coord, node_map<int>& node_type,
	const int i, const int k)
{
   edge_array<int> excess(T);
   calc_excess(T, excess, node_type);

   list<edge> L;  //list of edges that have been cut
   cut_edges (i, T, excess, L, k);
   node_array<int> node2Pnum(T); //maps a node to its piece
   int num_pieces; //number of pieces

   num_pieces = COMPONENTS(T, node2Pnum); //find the pieces
   if (num_pieces == 1) return TSPtour_len (T, node_coord);

   array<piece> P(num_pieces); //info for each piece
   list<int> plus_pieces;
   list<int> minus_pieces;
   get_pieces (T, L, excess, P, node2Pnum, plus_pieces, minus_pieces);
   match (T, node_coord, plus_pieces, minus_pieces, P);

   double l1 = ac_tour (T, node_coord, P, node2Pnum);
   double l2 = c_tour (T, node_coord, P, node2Pnum);

   return (min(l1, l2));
}

/*******************************************************************************
PROCEDURE TSPtour_len

Input:
Output:

*******************************************************************************/
double TSPtour_len (const UGRAPH<int, double>& T,
 	const node_map<point>& node_coord)
{
	double total_dist = 0;
	node u, v;
   u = T.first_node();
   while (u)
   {
   	v = T.succ_node(u);
      if (v != nil)
      	total_dist += dist(u, v, node_coord);
      else total_dist += dist(u, T.first_node(), node_coord);

      u = v;
   }

	return total_dist;
}

/*******************************************************************************
PROCEDURE next_ac_node

Input:  current node
Output: next node. returns distance travelled.

Helper function for ac_tour.
*******************************************************************************/
double next_ac_node (const UGRAPH<int, double>& T,
	const node_map<point>& node_coord,
	node& curr_node)
{
	node prev_node = curr_node;
   curr_node = T.succ_node(curr_node);
   if (curr_node == nil) curr_node = T.first_node();
   return dist(prev_node, curr_node, node_coord);
}

/*******************************************************************************
PROCEDURE next_c_node

Input:  current node
Output: next node. returns distance travelled.

Helper function for c_tour.
*******************************************************************************/
double next_c_node (const UGRAPH<int, double>& T,
	const node_map<point>& node_coord,
	node& curr_node)
{
	node prev_node = curr_node;
   curr_node = T.pred_node(curr_node);
   if (curr_node == nil) curr_node = T.last_node();
   return dist(prev_node, curr_node, node_coord);
}

/*******************************************************************************
PROCEDURE prev_ac_node

Input:  current node
Output: prev node in anticlocwise tour. returns distance travelled.

Helper function for ac_tour.
*******************************************************************************/
double prev_ac_node (const UGRAPH<int, double>& T,
	const node_map<point>& node_coord,
	node& curr_node)
{
	return next_c_node (T, node_coord, curr_node);
}

/*******************************************************************************
PROCEDURE prev_c_node

Input:  current node
Output: prev node in clockwise tour. returns distance travelled.

Helper function for c_tour.
*******************************************************************************/
double prev_c_node (const UGRAPH<int, double>& T,
	const node_map<point>& node_coord,
	node& curr_node)
{
	return next_ac_node (T, node_coord, curr_node);
}

/*******************************************************************************
PROCEDURE ac_tour

Input:
Output:

anti-clocwise traversal
*******************************************************************************/
double ac_tour (const UGRAPH<int, double>& T,
 	const node_map<point>& node_coord,
	array<piece> P,
   const node_array<int>& node2Pnum)
{

	double length = 0; //tour length

   int start_pnum = 1;
	int pnum = start_pnum; //piece number
   node old_node;
   node curr_node = P[pnum].start_node;

	do
   {
      switch (P[pnum].type)
      {
         case _zero:   	while (curr_node != P[pnum].end_node)
                        {
                        	//cout<<"zero piece\n";
                           length += next_ac_node(T, node_coord, curr_node);
                           
                        }
                        P[pnum].serviced = true;
                        break;
         case _plus:
         case _minus:	int matchPnum = P[pnum].match_piece;
         					bool start = true;
	                     do
         					{
                        	/****** Get the next node in the piece ******/
                        	if (!start)
                           {
                           	if (curr_node != P[pnum].end_node)
                           	{
                           		length += next_ac_node(T, node_coord, curr_node);
                           		
                           	}
                           }
                           else start = false;

                        	/****** Traverse the matched piece ******/
                           if (curr_node == P[pnum].match_node)
                           {
                           	if (P[matchPnum].serviced) break;
                              /* goto matched piece */
                              //cout<<"goto match piece\n";
                              old_node = curr_node;
                              curr_node = P[matchPnum].match_node;
                              length += dist(old_node, curr_node, node_coord);
                              /* goto start of piece */
                              //cout<<"start match piece\n";
                              while (curr_node != P[matchPnum].start_node)
                              {
                              	/* Get the previous node in the matched piece*/
                              	length += prev_ac_node(T, node_coord, curr_node);
                              }

                              /* traverse match piece */
                              //cout<<"traverse match piece\n";
                              while (curr_node != P[matchPnum].end_node)
                              {
                              	/* Get the next node in the matched piece*/
											length += next_ac_node(T, node_coord, curr_node);
                              }
                              P[matchPnum].serviced = true;

                              /* return to original piece */
		                        //cout<<"return\n";
                              while (curr_node != P[matchPnum].match_node)
                              {
                              	/* Get the previous node in the matched piece*/
                              	length += prev_ac_node(T, node_coord, curr_node);
                              }

                              old_node = curr_node;
                              curr_node = P[pnum].match_node;
                              length += dist(old_node, curr_node, node_coord);
                           }

                        }
         					while (curr_node != P[pnum].end_node);

                        P[pnum].serviced = true;
                        break;
      }

		length += next_ac_node(T, node_coord, curr_node);
      //cout<<"new piece\n";
      pnum = node2Pnum[curr_node];
   }
	while (pnum != start_pnum);

   return length;
}



/*******************************************************************************
PROCEDURE c_tour

Input:
Output:

clocwise traversal
*******************************************************************************/
double c_tour (const UGRAPH<int, double>& T,
 	const node_map<point>& node_coord,
	array<piece> P,
   const node_array<int>& node2Pnum)
{

	double length = 0; //tour length

   int start_pnum = 0;
	int pnum = start_pnum; //piece number
   node old_node;
   node curr_node = P[pnum].end_node;

	do
   {
      switch (P[pnum].type)
      {
         case _zero:   	while (curr_node != P[pnum].start_node)
                        {
                        	//cout<<"zero piece\n";
                           length += next_c_node(T, node_coord, curr_node);
                        }
                        P[pnum].serviced = true;
                        break;
         case _plus:
         case _minus:	int matchPnum = P[pnum].match_piece;
         					bool start = true;
	                     do
         					{
                        	/****** Get the next node in the piece ******/
                        	if (!start)
                           {
                           	if (curr_node != P[pnum].start_node)
                           	{
                           		length += next_c_node(T, node_coord, curr_node);
                           	}
                           }
                           else start = false;

                        	/****** Traverse the matched piece ******/
                           if (curr_node == P[pnum].match_node)
                           {
                           	if (P[matchPnum].serviced) break;
                              /* goto matched piece */
                              //cout<<"goto match piece\n";
                              old_node = curr_node;
                              curr_node = P[matchPnum].match_node;
                              length += dist(old_node, curr_node, node_coord);

                              /* goto start of piece */
                              //cout<<"start match piece\n";
                              while (curr_node != P[matchPnum].end_node)
                              {
                              	/* Get the previous node in the matched piece*/
                              	length += prev_c_node(T, node_coord, curr_node);
                              }

                              /* traverse match piece */
                              //cout<<"traverse match piece\n";
                              while (curr_node != P[matchPnum].start_node)
                              {
                              	/* Get the next node in the matched piece*/
											length += next_c_node(T, node_coord, curr_node);
                              }
                              P[matchPnum].serviced = true;
                              /* return to original piece */
                              //cout<<"return\n";
                              while (curr_node != P[matchPnum].match_node)
                              {
                              	/* Get the previous node in the matched piece*/
                              	length += prev_c_node(T, node_coord, curr_node);
                              }

                              old_node = curr_node;
                              curr_node = P[pnum].match_node;
                              length += dist(old_node, curr_node, node_coord);
                           }

                        }
         					while (curr_node != P[pnum].start_node);

                        P[pnum].serviced = true;
                        break;
      }

		length += next_c_node(T, node_coord, curr_node);
      //cout<<"new piece\n";
      pnum = node2Pnum[curr_node];
   }
	while (pnum != start_pnum);

   return length;
}

/*******************************************************************************
PROCEDURE match

Input:
Output:

*******************************************************************************/
void match (const UGRAPH<int, double>& T,
	const node_map<point>& node_coord,
	const list<int>& plusL, const list<int>& minusL,
   array<piece>& P)
{
   GRAPH<int, B_edge> B;
   /* max_len stores the length of the longes edge */
 	double max_len = create_BG (T, node_coord, plusL, minusL, P, B);

   /* partition nodes into two lists for mathcing algorithm */
   list<node> Bplus;
   list<node> Bminus;
   node v;
   forall_nodes (v, B)
   	if (P[B.inf(v)].type == _plus) Bplus.append(v);
      else Bminus.append(v);

   /* array of weights (after scaling) for matching algorithm */
   edge_array<int> weight(B);
   edge e;
   /* need to scale to make the max_match be a min_match */
   forall_edges(e, B) weight[e] = (int) rint((max_len - (B.inf(e)).length)*100);

   list<edge> result;
   result = MAX_WEIGHT_BIPARTITE_MATCHING(B, Bplus, Bminus, weight);

	/* add match info to piece*/
   forall (e, result) {
      int p1 = B.inf(B.source(e));
      int p2 = B.inf(B.target(e));
      if (P[p1].type == _plus) {
      	P[p1].match_node = (B.inf(e)).p_node;
         P[p1].match_length = (B.inf(e)).length;
         P[p1].match_piece = p2;
         P[p2].match_node = (B.inf(e)).m_node;
         P[p2].match_length = (B.inf(e)).length;
         P[p2].match_piece = p1;
      }
      else {
      	P[p1].match_node = (B.inf(e)).m_node;
         P[p1].match_length = (B.inf(e)).length;
         P[p1].match_piece = p2;
      	P[p2].match_node = (B.inf(e)).p_node;
         P[p2].match_length = (B.inf(e)).length;
         P[p2].match_piece = p1;
      }
   }
}

/*******************************************************************************
FUNCTION create_BG (used by PROCEDURE match)

Input: TSP-Graph, array of coordinates for each node, 2 lists of piece numbers
			one for plus pieces and one for minus pieces, Piece array
Output: Bi-partite graph.
			Returns the maximum length of an edge in the bipartite graph.

Creates a Bi-partite graph with a node for each piece and an edge representing
the shortest distance beteen two edges.
*******************************************************************************/
double create_BG (const UGRAPH<int, double>& T, const node_map<point>& node_coord,
	const list<int>& plusL, const list<int>& minusL,
   const array<piece>& P, GRAPH<int, B_edge>& B)
{

   /* array b_node maps a pice number to a node in B */
   array<node> Pnum2Bnode (P.size());
   int i;

   /* create nodes for each plus and minus piece  */
	forall (i, plusL) {
   	Pnum2Bnode[i] = B.new_node(i);
   }

   forall (i, minusL) {
   	Pnum2Bnode[i] = B.new_node(i);
   }
   double max_d = 0; //max edge weight in B
   int p, m;
	forall (p, plusL) { //for every plus piece
   	forall (m, minusL) { //find the min distance to every minus piece

   		node p_node = P[p].start_node;  //start of plus piece

         double min_d = -1; //min distance between the two pieces
         					//initialized to -1 to flag first time
         node p_min = nil, m_min = nil; //nodes with the min distance between
         											//them

         while (p_node != nil) { //iterate through all plus nodes of piece
         	node m_node = P[m].start_node;	//start of minus piece
         	while (m_node != nil) { //iterate through all minus nodes of piece

            	/* calculate distance between both nodes */
               double distance;
               distance = dist(p_node, m_node, node_coord);
               if ((distance < min_d) || (min_d == -1)) {
               	min_d = distance;
                  p_min = p_node;
                  m_min = m_node;
               }


               /* get next minus node */
               if (m_node != P[m].end_node) {
               	m_node = T.succ_node(m_node);
                  /* cycle through vertex set */
      				if (m_node == nil) m_node = T.first_node();
               }
               else m_node = nil;

            }//while m_node

            /* get next plus node */
            if (p_node != P[p].end_node) {
               p_node = T.succ_node(p_node);
               /* cycle through vertex set */
               if (p_node == nil) p_node = T.first_node();
            }
            else p_node = nil;
         }//while p_node

         /* create edge in bipartite graph and set its weight */
         B_edge edge_info;
         edge_info.p_node = p_min;
         edge_info.m_node = m_min;
         edge_info.length = min_d;
         B.new_edge(Pnum2Bnode[p], Pnum2Bnode[m], edge_info);

         if (min_d > max_d) max_d = min_d; //maximum edge length in B

      }//forall m
   }//forall p


   return max_d;
} //create_BG



/*******************************************************************************
PROCEDURE calc_excess

Input: tsp solution graph, array to hold excess values for each edge
Output: Graph with excess function calculated for each edge.

	for each edge, e, calculates
		EXCESS(e) = PEGS(e) - SLOTS(e)
	where
		PEGS(e) = number of pegs visited before e is traversed
  		SLOTS(e) = number of slots visited before e is traversed
*******************************************************************************/
void calc_excess(UGRAPH<int, double>& T, edge_array<int>& excess,
	const node_map<int>& node_type)
{
	int pegs = 0, slots = 0;
   edge e = T.first_edge();
   while (e != nil) {
      if (node_type(T.source(e)) == _peg) pegs++;
      else slots++;
   	excess[e] = (pegs - slots);
      e = T.succ_edge(e);
   }

}


/*******************************************************************************
PROCEDURE cut_edges

Input: i, tsp solution, array of excess values for each edge
Output: list of edges cut
*******************************************************************************/
void cut_edges (const int i, UGRAPH<int, double>& T, const edge_array<int> excess,
	list<edge>& L, const int k)
{
	edge e;

   /* if  i = excess(e) mod k/2, select edge for cutting */
	forall_edges(e, T)
   	if (i == excess[e] % (k/2))
      	L.append(e);

	/* cut edges */
	forall(e, L) T.hide_edge(e);
   //forall(e, L) T.del_edge(e);
}


/*******************************************************************************
PROCEDURE get_pieces

Input: graph, list of cut edges, array identifying the piece each node
			belongs to, array of excess values for each edge
Output: array of information (indexed by piece number) for each piece
*******************************************************************************/
void get_pieces (const UGRAPH<int, double>& T, const list<edge>& L,
	const edge_array<int> excess,
	array<piece>& P, const node_array<int>& node2Pnum, list<int>& plusL,
   list<int>& minusL)
{
   if (!L.empty()) {

   	list_item it;
      edge prev_e, next_e;	//previous and next cut edge of piece
      node start_node;	//of piece
      int n;

      forall_items (it, L) {
      	/* get cut edges bounding piece */
         prev_e = L.contents(it);
      	next_e = L.contents(L.cyclic_succ(it));

         start_node = T.target(prev_e);
         n = node2Pnum[start_node];
         P[n].start_node = start_node;
         P[n].end_node = T.source(next_e);


         /* Get the length of the piece */
         /*
         edge curr_edge = prev_e;
         if (curr_edge != T.last_edge())
         	curr_edge = T.succ_edge(curr_edge);
         else curr_edge = T.first_edge();

         int length = 0; //length of piece
         while (curr_edge != next_e)
			{
         	length += T.inf(curr_edge);
         	if (curr_edge != T.last_edge())
      	   	curr_edge = T.succ_edge(curr_edge);
   	      else curr_edge = T.first_edge();
         }
         P[n].length = length;
         */

         if (excess[prev_e] == excess[next_e]) {
         	/* zero piece */
            P[n].type = _zero;
         }
         else if (excess[prev_e] < excess[next_e]) {
         	/* plus piece */
            P[n].type = _plus;
				plusL.append(n);
         }
         else if (excess[prev_e] > excess[next_e]) {
         	/* minus piece */
            P[n].type = _minus;
            minusL.append(n);
         }

      }

   }

}






/*******************************************************************************
CLASS piece
*******************************************************************************/

piece::piece()
{
  	type = 0;
   start_node = nil;
   end_node = nil;
   length = 0;
   match_piece = -1;
   match_node = nil;
   match_length = -1;
   serviced = 0;
}

piece::piece (const piece& P)
{
	type = P.type;
   start_node = P.start_node;
   end_node = P.end_node;
   length = P.length;
   match_piece = P.match_piece;
   match_node = P.match_node;
   match_length = P.match_length;
   serviced = P.serviced;
}

piece& piece::operator= (const piece& P)
{
	if (&P != this) {
      type = P.type;
      start_node = P.start_node;
      end_node = P.end_node;
      length = P.length;
      match_piece = P.match_piece;
	   match_node = P.match_node;
   	match_length = P.match_length;
      serviced = P.serviced;
   }
   return *this;
}

istream& operator>> (istream& IS, piece& P)
{
	IS>>P.type;
   IS>>P.start_node;
   IS>>P.end_node;
   IS>>P.length;
   IS>>P.match_piece;
   IS>>P.match_node;
   IS>>P.match_length;
   IS>>P.serviced;
   return IS;
}

ostream& operator<< (ostream& OS, const piece& P)
{
   OS<<"start_node "<<P.start_node<<'\n';
   OS<<"start_node "<<P.end_node<<'\n';
   OS<<"length "<<P.length<<'\n';
  	OS<<"type "<<P.type<<'\n';
   OS<<"match_piece "<<P.match_piece<<'\n';
   OS<<"match_node "<<P.match_node<<'\n';
   OS<<"match_length "<<P.match_length<<'\n';
   OS<<"serviced "<<P.serviced<<'\n';
   return OS;
}

/**********************************************